#import "Vector3f.h"
#import "Vector4f.h"
#import "DualQuaternion.h"

@interface BONE : NSObject {
	
    NSString       *boneBoneName;
    GLushort        boneIndex;
    GLushort        boneParentIndex;
    NSMutableArray *boneChildIndices;
    NSMutableArray *boneAnimations;                 // Local-space array of per-frame (Matrix4f) animation transforms, either imported with 'io_scene_p2' or synthesized (See 1).
    GLushort        boneAnimationLength;            // Number of animation frames for the bone; can differ between bones in an Armature, a BVH animation, and so on.

    DualQuaternion *boneLsBindTransform;            // Local-space transform of bone from Parent bone to original (un-posed) bind position
    DualQuaternion *boneMsInverseBindTransform;     // Model-space transform of bone from original (un-posed) bind position to the Armature's origin
    DualQuaternion *boneLsAnimatedPoseTransform;    // Local-space transformation matrix of bone from original bind position to an animated pose position; interpolated every frame
    DualQuaternion *boneMsAnimatedPoseTransform;    // Model-space transform (as a Dual Quaternion) of bone's Local-space animated pose position (above); recalculated every frame
    
    Vector3f       *boneLocation;                   // Local-space decompositions of localBindTransform's Location, Rotation and Scale values from Blender export add-on
    Vector4f       *boneRotation;
    Vector3f       *boneScale;
}

-(id) init  :(GLushort)new_Index  :(NSString*)new_Name  :(GLushort)new_ParentIndex;

-(NSString*)        getBoneName;
-(GLushort)         getIndex;
-(GLushort)         getParentIndex;
-(NSMutableArray*)  getChildIndices;
-(NSMutableArray*)  getAnimations;
-(GLushort)         getAnimationLength;
-(DualQuaternion*)  getLsBindTransform;
-(DualQuaternion*)  getMsInverseBindTransform;
-(DualQuaternion*)  getLsAnimatedPoseTransform;
-(DualQuaternion*)  getMsAnimatedPoseTransform;
-(Vector3f*)        getLocation;
-(Vector4f*)        getRotation;
-(Vector3f*)        getScale;

-(void) setChildIndices             :(NSMutableArray*)new_ChildIndices;
-(void) setAnimations               :(NSMutableArray*)new_Animations;
-(void) setAnimationLength          :(GLushort)new_AnimationLength;
-(void) setLsBindTransform          :(DualQuaternion*)new_LocalBindTransform;
-(void) setMsInverseBindTransform   :(DualQuaternion*)new_ModelInverseBindTransform;
-(void) setLsAnimatedPoseTransform  :(DualQuaternion*)new_LocalAnimatedPoseTransform;
-(void) setMsAnimatedPoseTransform  :(DualQuaternion*)new_ModelAnimatedPoseTransform;
-(void) setLocation                 :(Vector3f*)new_Location;
-(void) setRotation                 :(Vector4f*)new_Rotation;
-(void) setScale                    :(Vector3f*)new_Scale;

@end

//  (1) This boneAnimations array is only used for ArmatureList (Blender-imported) and ActionList (synthesized) animations;
//      BVHList processing only requires BoneKeyFrames for animation.
