#ifndef Constants_h
#define Constants_h

#define MAX_BONES           66              // Maximum Bones - Must NOT be reached (or exceeded) and MUST match Shader.vert definition
#define MAX_WEIGHTS         4               // A maximum of 4 bone weights with which to deform a given vertex (i.e. to fit in a vec4)
#define MAX_NAMESIZE        63+1            // Maximum length of string (including terminating null) for various names, e.g. filenames
#define MAX_DEBUG_FRAMES    1               // Maximum number of frames to log debug information  (pose transform, EulerXYZ, position)

#define dummyTexelName (char*)"dummy1x1Texel"

enum AnimationTypes { typeArmature, typeAction, typeBVH, nAnimationTypes };

enum {                                      // P-Squared (.p2) meshes are derived from .3ds format, adding 32-bit CHUNK_VERTFLAGS, new CHUNK_ARMATURE_BONES, et al.
    CHUNK_M3D_VERSION           = 0x0002,
    CHUNK_COLOR_F			    = 0x0010,
    CHUNK_COLOR_24			    = 0x0011,
    CHUNK_LIN_COLOR_24		    = 0x0012,
    CHUNK_LIN_COLOR_F           = 0x0013,
    CHUNK_INT_PERCENTAGE	    = 0x0030,
    CHUNK_FLOAT_PERCENTAGE      = 0x0031,
    CHUNK_MASTER_SCALE          = 0x0100,
    CHUNK_SOLID_BGND		    = 0x1200,
    CHUNK_USE_SOLID_BGND        = 0x1201,
    CHUNK_AMBIENT_LIGHT		    = 0x2100,
    CHUNK_MDATA                 = 0x3D3D,
    CHUNK_MESH_VERSION          = 0x3D3E,
    CHUNK_NAMED_OBJECT		    = 0x4000,
    CHUNK_N_TRI_OBJECT		    = 0x4100,
    CHUNK_POINT_ARRAY           = 0x4110,
    CHUNK_VERTFLAGS             = 0x4111,   // 32-bit flags encodes Blender Vertex Group IDs/weights; diverges from ancient .3ds file format specifications
    CHUNK_ARMATURE_BONES        = 0x4112,   // CUSTOM chunk encodes Blender Armature Bone indexes, names, locations, rotations, scales, animation keyframes...
    CHUNK_FACE_ARRAY		    = 0x4120,
    CHUNK_MSH_MAT_GROUP		    = 0x4130,
    CHUNK_TEX_VERTS			    = 0x4140,
    CHUNK_SMOOTH_GROUP		    = 0x4150,
    CHUNK_MESH_MATRIX		    = 0x4160,
    CHUNK_N_DIRECT_LIGHT        = 0x4600,
    CHUNK_DL_SPOTLIGHT		    = 0x4610,
    CHUNK_N_CAMERA			    = 0x4700,
    CHUNK_M3DMAGIC              = 0x4D4D,
    CHUNK_MAT_ENTRY			    = 0xAFFF,
    CHUNK_MAT_NAME			    = 0xA000,
    CHUNK_MAT_AMBIENT		    = 0xA010,
    CHUNK_MAT_DIFFUSE		    = 0xA020,
    CHUNK_MAT_SPECULAR		    = 0xA030,
    CHUNK_MAT_SHININESS		    = 0xA040,
    CHUNK_MAT_SHIN2PCT  	    = 0xA041,
    CHUNK_MAT_SHIN3PCT          = 0XA042,
    CHUNK_MAT_TRANSPARENCY	    = 0xA050,
    CHUNK_MAT_TWO_SIDE		    = 0xA081,
    CHUNK_MAT_XPFALLIN		    = 0xA08A,
    CHUNK_MAT_SHADING           = 0xA100,
    CHUNK_MAT_TEXMAP		    = 0xA200,
    CHUNK_MAT_BUMPMAP		    = 0xA230,
    CHUNK_MAT_BUMP_PERCENT      = 0xA252,
    CHUNK_MAT_MAPNAME		    = 0xA300,
    CHUNK_MAT_TEX2MAP           = 0xA33A,
    CHUNK_MAT_MAP_TILING	    = 0xA351,
    CHUNK_MAT_MAP_TEXBLUR       = 0xA353,
    CHUNK_MAT_MAP_USCALE        = 0xA354,
    CHUNK_MAT_MAP_VSCALE        = 0xA356,
    CHUNK_MAT_MAP_UOFFSET       = 0xA358,
    CHUNK_MAT_MAP_VOFFSET       = 0xA35A,
    CHUNK_MAT_MAP_ANG           = 0xA35C,
    CHUNK_MAT_MAP_COL1          = 0xA360,
    CHUNK_MAT_MAP_COL2          = 0xA362,
    CHUNK_MAT_MAP_RCOL          = 0xA364,
    CHUNK_MAT_MAP_GCOL          = 0xA366,
    CHUNK_MAT_MAP_BCOL          = 0xA368,
    CHUNK_KFDATA			    = 0xB000,
    CHUNK_OBJECT_NODE_TAG       = 0xB002,
    CHUNK_CAMERA_NODE_TAG       = 0xB003,
    CHUNK_TARGET_NODE_TAG       = 0xB004,
    CHUNK_LIGHT_NODE_TAG        = 0xB005,
    CHUNK_LTARGET_NODE_TAG      = 0xB006,
    CHUNK_SPOT_NODE_TAG         = 0xB007,
    CHUNK_KFSEG			        = 0xB008,
    CHUNK_KFCURTIME             = 0xB009,
    CHUNK_KFHDR                 = 0xB00A,
    CHUNK_NODE_HDR              = 0xB010,
    CHUNK_INSTANCE_NAME         = 0xB011,
    CHUNK_OBJECT_PIVOT          = 0xB013,
    CHUNK_OBJECT_BOUNDBOX       = 0xB014,
    CHUNK_OBJECT_MORPH_SMOOTH   = 0xB015,
    CHUNK_POS_TRACK_TAG         = 0xB020,
    CHUNK_ROT_TRACK_TAG         = 0xB021,
    CHUNK_SCL_TRACK_TAG         = 0xB022,
    CHUNK_FOV_TRACK_TAG         = 0xB023,
    CHUNK_ROLL_TRACK_TAG        = 0xB024,
    CHUNK_COL_TRACK_TAG         = 0xB025,
    CHUNK_HOTSPOT_TRACK_TAG     = 0xB027,
    CHUNK_FALLOFF_TRACK_TAG     = 0xB028,
    CHUNK_OBJECT_NODE_ID        = 0xB030,
    CHUNK_OBJECT_PARENT_NAME    = 0x80F0
};

#endif /* Constants_h */
