#define GLM_ENABLE_EXPERIMENTAL
#import <glm/gtc/type_ptr.hpp>
#import <glm/gtx/dual_quaternion.hpp>
#import <eigen3/Eigen/Dense>

#import "Vector3f.h"
#import "Matrix4f.h"
#import "Quaternion.h"

@interface DualQuaternion : NSObject {
    
    Quaternion  *m_real;            // Rotation part
    Quaternion  *m_dual;            // Translation part
}

-(id)   init                       :(Quaternion*)real  :(Quaternion*)dual;
-(id)   initFromGLM                :(glm::mat4)in_glmMatrix;
-(id)   initFromMatrix             :(Matrix4f*)in_Matrix;
-(id)   initFromAngleAxis          :(float)in_AngleDegrees  :(Vector3f*)in_Axes;
-(NSString*) toString;

+(Quaternion*)     getRotation     :(DualQuaternion*)dq;
+(Vector3f*)       getTranslation  :(DualQuaternion*)dq;
+(Quaternion*)     getReal         :(DualQuaternion*)dq;
+(Quaternion*)     getDual         :(DualQuaternion*)dq;
+(DualQuaternion*) normalize       :(DualQuaternion*)dq;
+(DualQuaternion*) conjugate       :(DualQuaternion*)dq;
+(DualQuaternion*) multiply        :(DualQuaternion*)lhs  :(DualQuaternion*)rhs;
+(DualQuaternion*) interpolate     :(DualQuaternion*)fromDQ  :(DualQuaternion*)toDQ  :(float)progression;
+(glm::dualquat)   toGLMdualquat   :(DualQuaternion*)dq;

@end
