#import "File_Routines.h"

@implementation MeshP2 (File_Routines)

-(void) Dread  :(void*)dest  :(int)len {                                // Reads len bytes from InFile into the memory pointed to by *dest
    
    if (fread (dest, len, 1, InFile) != 1) {
        
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->Dread: Failed on fread with errno: %d (%s)", errno, strerror (errno)];
        return;
    }
    
	if (CFByteOrderGetCurrent() == CFByteOrderLittleEndian) return;	    // If on an Intel CPU, no conversion needed
	
    if (len == 2) {                                                     // Otherwise, convert little endian .p2 data into big endian format
        
		*(UInt16*) dest = CFSwapInt16 (*(UInt16*)dest);
		return;
    }
	
    if (len == 4) {
        
		*(UInt32*) dest = CFSwapInt32 (*(UInt32*)dest);
		return;
    }
}

-(void) Dsetpos  :(int) pos {
    
    if (fseek (InFile, (long)pos, SEEK_SET) != 0) {
        
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->Dsetpos: Failed on fseek with errno: %d (%s)", errno, strerror (errno)];
    }
}

-(int)  Dgetpos {
    
    long pos;
  	
	if ((pos = ftell (InFile)) == -1L) {
		
		[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->Dsetpos failed on ftell with errno: %d (%s)", errno, strerror (errno)];
	};
		
    return (int) pos;
}

-(bool) ReadScene  :(FILE*)ptr  :(int)size {
    
    ushort  ident;
    int     len, pc;
    
    InFile = ptr;
    
    pc = [self Dgetpos];
    
    [self Dread  :&ident  :sizeof(ident)];
    [self Dread  :&len    :sizeof(len)];
    
    if (ident != CHUNK_M3DMAGIC) {
        
		[MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadScene: Mesh file is not in P2 format."];
		return false;
    }
    
    [self ReadMainBlocks  :(pc+len)];
    
    return true;
}

-(void) ReadMainBlocks  :(int)p {
    
    ushort  ident;
    int     len, pc;
    
    while ((pc = [self Dgetpos]) < p) {
        
		[self Dread  :&ident  :sizeof(ident)];
		[self Dread  :&len    :sizeof(len)];
		
		switch (ident) {
                
            case CHUNK_M3D_VERSION:

                [self Dsetpos  :(pc+len)];
                break;
                
			case CHUNK_MDATA:
                
				[self ReadObjMeshBlocks  :(pc+len)];
                break;
                
			case CHUNK_KFDATA:

                [self ReadKFDataBlocks  :(pc+len)];
                break;
                
			default:
                
				[self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadMainBlocks: *** Unhandled Chunk IGNORED *** Ident:0x%X, Length: %d", ident, len];
        }
    }
}

-(void) ReadObjMeshBlocks  :(int)p {

    ushort  ident;
    int     len, pc;
    
    while ((pc = [self Dgetpos]) < p) {

		[self Dread  :&ident  :sizeof(ident)];
		[self Dread  :&len    :sizeof(len)];
        
        switch (ident) {
                
            case CHUNK_NAMED_OBJECT:
                
                [self ReadObjBlocks  :(pc+len)];
                break;
                
            case CHUNK_MAT_ENTRY:
                
                [self ReadMaterialBlocks  :(pc+len)];
                break;

            case CHUNK_MASTER_SCALE:
                
                [self ReadMasterScale  :(pc+len)];
                break;
                
            case CHUNK_MESH_VERSION:
                
                [self Dsetpos  :(pc+len)];
                break;
                
            case CHUNK_AMBIENT_LIGHT:
                
                [self Dsetpos  :(pc+len)];
                break;

            case CHUNK_SOLID_BGND:
                
                [self Dsetpos  :(pc+len)];
                break;
                
            case CHUNK_USE_SOLID_BGND:
                
                [self Dsetpos  :(pc+len)];
                break;
                
            default:
                
                [self Dsetpos  :(pc+len)];

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjMeshBlocks: *** Unhandled Chunk IGNORED *** Ident:0x%X, Length: %d", ident, len];
        }
    }
}

-(void) ReadObjBlocks  :(int)p {

    ushort  ident;
    int     len, pc;
    char    filename[MAX_NAMESIZE] = {0};
      
    [self ReadCString  :filename  :sizeof(filename)];   // Name of the original mesh
    
    while ((pc = [self Dgetpos]) < p) {

        [self Dread  :&ident  :sizeof(ident)];
        [self Dread  :&len    :sizeof(len)];
        
        switch (ident) {

            case CHUNK_N_TRI_OBJECT:
                
                [self ReadTriMeshBlocks  :pc+len  :filename];
                break;
                
            case CHUNK_N_DIRECT_LIGHT:                  // Blender lights and cameras are ignored here...
                
                [self Dsetpos  :(pc+len)];
                break;
                
            case CHUNK_N_CAMERA:                        // ... as they're managed by MeshP2's calling app
                
                [self Dsetpos  :(pc+len)];
                break;
                
            default:

                [self Dsetpos  :(pc+len)];

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjBlocks: *** UNHANDLED IDENT *** Ident:0x%X, Length:%d", ident, len];
        }
    }
    
    [self Dsetpos  :p];
}

-(void) ReadTriMeshBlocks  :(int)p  :(char*)name {

    ushort  ident;
    ushort  numFaces = 0;
    bool    found_facematerial = false;
    int     len, pc;

    P2MeshObj *meshObj = [Scene addMeshObj];                // Add a new P2MeshObj to the Scene (Mixamo meshes can have >1)
    
    [meshObj setMeshName  :name];

    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTriMeshBlocks: '%s'", name];

    while ((pc = [self Dgetpos]) < p) {

        [self Dread  :&ident  :sizeof(ident)];
        [self Dread  :&len    :sizeof(len)];
        
        switch (ident) {

            case CHUNK_POINT_ARRAY:
                
                [self ReadVertList  :meshObj];
                break;

            case CHUNK_VERTFLAGS:
                
                [self ReadVertFlags  :meshObj];
                break;

            case CHUNK_ARMATURE_BONES:
                
                if (![self ReadArmatureBones  :meshObj]) {      // If too many bones, or other error...
                    
                    [self Dsetpos  :(pc+len)];                  // Skip over Armature block entirely
                }
                break;
                
            case CHUNK_FACE_ARRAY:
                
                numFaces = [self ReadFaceList  :meshObj];
                break;
                
            case CHUNK_TEX_VERTS:
                
                [self ReadUVMapList  :meshObj];
                break;

            case CHUNK_MSH_MAT_GROUP:
                
                [self ReadFaceMaterial  :meshObj];
                
                found_facematerial = true;
                break;
                
            case CHUNK_MESH_MATRIX:
                
                [self ReadTransformMatrix  :meshObj];
                break;
                
            case CHUNK_SMOOTH_GROUP:                // Unsaved and unused

                [self Dsetpos  :(pc+len)];
                break;
                
            default:

                [self Dsetpos  :p];                 // Bypass illegal TriMeshBlock
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTriMeshBlocks: *** ILLEGAL TriMeshBlock BYPASSED *** Ident:0x%X, Length: %d", ident, len];
        }
    }
    
    if    (!found_facematerial) [self WriteDefaultMaterial  :meshObj  :numFaces];   // Couldn't find any face materials, so make one up
}

-(void) ReadVertList  :(P2MeshObj*)meshObj {

    ushort      nv;
    float       x, y, z;

    [self Dread  :&nv  :sizeof(nv)];
    
    for (int n=0; n < nv; n++) {
       
        [self Dread  :&x  :sizeof(x)];
        [self Dread  :&y  :sizeof(y)];
        [self Dread  :&z  :sizeof(z)];

        Vector4f *position = [[Vector4f alloc] init  :x  :y  :z  W:1.0];

        P2Vertex *vertex_entry = [meshObj addVert];

        [vertex_entry setX  :[position getX]];
        [vertex_entry setY  :[position getY]];
        [vertex_entry setZ  :[position getZ]];
    }
}

-(void) ReadVertFlags  :(P2MeshObj*)meshObj {
/*
    P2 mesh format is a derivation of the obsolete .3DS 'specification' from http://dzzd.net/3DSChunkDefinitions.html diverging where
    P2 vertex flags are 32-bit uints, not ushorts, in the expectation they were exported by our custom Blender add-on: 'io_scene_p2'.
*/
    uint32  nFlags;
    uint32  flag;
    
    [self Dread  :&nFlags  :sizeof(nFlags)];
    
    for (int n=0; n < nFlags; n++) {

        P2VertexFlag *vertex_flag_entry = [meshObj addVertFlag];
        
        [self Dread  :&flag  :sizeof(flag)];
        
        [vertex_flag_entry setFlag  :flag];
    }
}

-(bool) ReadArmatureBones  :(P2MeshObj*)meshObj {

//  Diverging from the 'specification' at http://dzzd.net/3DSChunkDefinitions.html we task an unused 3DS chunk ID to read Blender's resting-state Armature Bones,
//  animation transformation matrices (if any) and all other Blender bone data via packed strings (i.e. struct.pack) exported by our custom 'io_scene_p2' add-on.

    ushort          nBones;
    ushort          boneIndex;
    ushort          parentIndex;
    ushort          nChildren, childIndex;
    ushort          nAnimationTransforms;
    
    char            boneName[MAX_NAMESIZE]      = {0};                      // Null terminated character array of bone's name
    unsigned char   localBindTransformBytes[64] = {0};                      // Raw bytes of Local-space bind position transform
    unsigned char   importedAnimationBytes[64]  = {0};                      // Raw bytes of Local-space bone animation transform
    unsigned char   boneNameSize;

    [self Dread  :&nBones  :sizeof(nBones)];

    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->File_Routines->ReadArmatureBones: nBones:%d, limit:%d", nBones, MAX_BONES];

    if (nBones >= MAX_BONES) {                                              // Has the bone count reached the bone count guardrail?

        if (!tooManyBonesWarning) {                                         // Issue a (one-time) warning and ignore the Armature
            
            tooManyBonesWarning = true;
            
            NSAlert *alert = [NSAlert new];
            
            [alert setAlertStyle       :NSAlertStyleWarning];
            [alert setMessageText      :@"Armature will be BYPASSED!"];
            [alert setInformativeText  :[NSString stringWithFormat  :@"Armature has %d defined bones, reaching MAX_BONES limit of %d", nBones, MAX_BONES]];
            [alert addButtonWithTitle  :@"OK"];
            [alert runModal];
        }

        return false;                                                       // Tell caller to skip over the block
    }
    
    for (ushort i=0; i < nBones; i++) {                                     // Otherwise, for all available bones:
        
        P2ArmatureBone *bone_entry = [meshObj addArmatureBone];             // Add a P2ArmatureBone to P2MeshObj's armatureList

        NSMutableArray *children    = [NSMutableArray new];                 // Instantiate a new child bone array
        NSMutableArray *animations  = [NSMutableArray new];                 // Instantiate a new animation transforms array

        Vector3f *boneLocation      = [Vector3f new];                       // Instantiate a new Location, Rotation & Scale vector
        Vector4f *boneRotation      = [Vector4f new];
        Vector3f *boneScale         = [Vector3f new];
        
        [self Dread  :&boneIndex    :sizeof(boneIndex)];                    // Bone Index (aka Vertex Group index)
        [self Dread  :&parentIndex  :sizeof(parentIndex)];                  // Parent bone index, or 0 for Root
        [self Dread  :&nChildren    :sizeof(nChildren)];                    // Count of this bone's children
        
        [bone_entry       setIndex  :boneIndex];                            // Set P2ArmatureBone boneIndex
        [bone_entry setParentIndex  :parentIndex];                          //  "         "       parentIndex
        
        assert (boneIndex == i);                                            // Trap any Blender export issues

        if (nChildren) {                                                    // If this bone has children:
            
            for (ushort j=0; j < nChildren; j++) {                          // Propagate child indices
                
                [self Dread  :&childIndex  :sizeof(childIndex)];
                
                [children addObject  :[[NSNumber alloc] initWithUnsignedShort  :childIndex]];
            }
        }

        [bone_entry setChildIndices  :children];                            // Set P2ArmatureBone entry

        [self Dread  :&boneNameSize  :sizeof(boneNameSize)];                // Length of the bone name string

        assert (boneNameSize <= MAX_NAMESIZE);                              // Verify bone name size is supportable
        
        [self ReadCString  :boneName  :boneNameSize];                       // Read all expected bone name characters

        [bone_entry setName  :[[NSString alloc] initWithBytes  :boneName length:boneNameSize encoding:NSASCIIStringEncoding]];

        [self Dread  :[boneLocation getXptr]  :sizeof([boneLocation getX])];    // Read bone's (Local-space) bind location vector
        [self Dread  :[boneLocation getYptr]  :sizeof([boneLocation getY])];
        [self Dread  :[boneLocation getZptr]  :sizeof([boneLocation getZ])];

        [self Dread  :[boneRotation getWptr]  :sizeof([boneRotation getW])];    // Read W value first ('real' part of quaternion)
        [self Dread  :[boneRotation getXptr]  :sizeof([boneRotation getX])];    // Read X, Y and Z values ('imaginary' parts)
        [self Dread  :[boneRotation getYptr]  :sizeof([boneRotation getY])];
        [self Dread  :[boneRotation getZptr]  :sizeof([boneRotation getZ])];
        
        [self Dread  :[boneScale getXptr]     :sizeof([boneScale getX])];       // Read bone's scaling vector
        [self Dread  :[boneScale getYptr]     :sizeof([boneScale getY])];
        [self Dread  :[boneScale getZptr]     :sizeof([boneScale getZ])];

        [bone_entry setLocation  :boneLocation];                                // Set P2ArmatureBone entries
        [bone_entry setRotation  :boneRotation];
        [bone_entry    setScale  :boneScale];
        
        [self Dread  :&localBindTransformBytes  :sizeof(localBindTransformBytes)];  // Read Bind Position (i.e. unanimated, Local-space resting pose)

        [bone_entry setLsBindTransform  :[[Matrix4f new] rmMatrixFromBytes  :(float*)localBindTransformBytes]]; // Local bind transform to Matrix4f

        [self Dread  :&nAnimationTransforms  :sizeof(nAnimationTransforms)];        // Read count of this bones animation transformation matrices

        if (nAnimationTransforms) {                                                 // If there are animation transform matrices for this bone:
            
            for (ushort k=0; k < nAnimationTransforms; k++) {
                
                [self Dread  :&importedAnimationBytes  :sizeof(importedAnimationBytes)];    // Read the raw bytes of animation transform matrix

                [animations addObject  :[[Matrix4f new] rmMatrixFromBytes  :(float*)importedAnimationBytes]];   // Add to animation array
            }
        }
        
        [bone_entry      setAnimations  :animations];                           // Set imported animations in P2ArmatureBone
        [bone_entry setAnimationLength  :nAnimationTransforms];                 // Reuse animation transform count as length
    }
    
    return true;
}

-(ushort) ReadFaceList  :(P2MeshObj*)meshObj {

    ushort  nf;
    ushort  swap;
    ushort  p0, p1, p2, flags;
    
    [self Dread  :&nf  :sizeof(nf)];
    
    for (int n=0; n < nf; n++) {

        P2Face *face_entry = [meshObj addFace];
        
        [self Dread  :&p0  :sizeof(p0)];
        [self Dread  :&p1  :sizeof(p1)];
        [self Dread  :&p2  :sizeof(p2)];
        
        [self Dread  :&flags  :sizeof(flags)];
        
        if (flags == 0x07) {                        // 0x07 means AB, BC, and CA edges are wound in reverse order, so swap the last two vertices
            
            swap    = p1;
            p1      = p2;
            p2      = swap;
        }
        
        [face_entry setP0  :p0];
        [face_entry setP1  :p1];
        [face_entry setP2  :p2];
        [face_entry setFlags  :flags];
    }
    
    return (nf);
}

-(void) ReadUVMapList  :(P2MeshObj*)meshObj {

    ushort  nv;
    float   u, v;
    
    [self Dread  :&nv  :sizeof(nv)];
    
    for (int n=0; n < nv; n++) {

        P2Map *map_entry = [meshObj addMap];
        
        [self Dread  :&u  :sizeof(float)]; [map_entry setU  :u];
        [self Dread  :&v  :sizeof(float)]; [map_entry setV  :v];
    }
}

-(void) ReadFaceMaterial  :(P2MeshObj*)meshObj {

    ushort  nf;
    ushort  facevalue;
    char    name[MAX_NAMESIZE] = {0};
    
    [self ReadCString  :name  :MAX_NAMESIZE];
    
    P2MaterialFacesList *matFaceList_entry = [meshObj addMatList];
    
    [matFaceList_entry setName  :name];
    
    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadFaceMaterial: '%s'", name];
    
    [self Dread  :&nf  :sizeof(nf)];                    // Then the number of faces
    
    for (int n=0; n < nf; n++) {                        // Bind the face values, which are meshObj->faceList array indexes, to this material name

        [self Dread  :&facevalue  :sizeof(facevalue)];
        
        [matFaceList_entry addFace  :facevalue];
    }
}

-(void) ReadTransformMatrix  :(P2MeshObj*)meshObj {

    float   transformBytes[16];

    for (int i=0; i < 16; i++) {
        
        [self Dread  :&transformBytes[i]  :sizeof(float)];
    }
    
    [meshObj setTransformMatrix  :transformBytes];                              // Set mesh object's transformation matrix  //FIXME: CLARIFY .vs. MASTER_SCALE
    
    if (P2_Logging_Enabled) {
            
            glm::mat4 glmMat4Transform  = glm::make_mat4 (transformBytes);

            NSString *strMat4Transform  = [NSString stringWithUTF8String  :glm::to_string (glmMat4Transform).c_str()];
            
            NSString *strFormat         = [NSString stringWithUTF8String  :"MeshP2->ReadTransformMatrix:\n\tRowMajorMat4:\t%@\n"];
            
            [MeshP2 P2Log  :P2_Logging_Enabled  :strFormat, strMat4Transform];
    }
}

-(void) WriteDefaultMaterial  :(P2MeshObj*)meshObj  :(ushort)numFaces {

    P2MaterialFacesList *matFaceList_entry = [meshObj addMatList];
    
    [matFaceList_entry setName  :(char*)"WG_DEFAULT"];  // Add a (legacy?) material name
    
    for (int n=0; n < numFaces; n++) {                  // Bind all faces to this default material

        [matFaceList_entry addFace  :n];
    }
}

-(void) ReadMaterialBlocks  :(int)p {

    ushort  ident;
    int     len, pc;
    
    P2Material *materialObject = [Scene addMaterial];
    
    while ((pc = [self Dgetpos]) < p) {

		[self Dread  :&ident  :sizeof(ident)];
		[self Dread  :&len    :sizeof(len)];
        
		switch (ident) {

			case CHUNK_MAT_NAME:
                
				[self ReadMatNameChunk  :materialObject];
				break;
                
			case CHUNK_MAT_AMBIENT:
                
				[self ReadShadingChunk  :(pc+len)  :[materialObject getAmbient]];
				break;
                
			case CHUNK_MAT_DIFFUSE:
                
				[self ReadShadingChunk  :(pc+len)  :[materialObject getDiffuse]];
				break;
                
			case CHUNK_MAT_SPECULAR:
                
				[self ReadShadingChunk  :(pc+len)  :[materialObject getSpecular]];
				break;
                    
            case CHUNK_MAT_SHININESS:
                
                [self ReadShininessChunk  :(pc+len)  :[materialObject getShininess]];
                break;
                
            case CHUNK_MAT_SHIN2PCT:
                
                [self Dsetpos  :(pc+len)];      // Currently unused
                break;
                
            case CHUNK_MAT_SHIN3PCT:
                
                [self Dsetpos  :(pc+len)];      // Currently unused
                break;

            case CHUNK_MAT_TRANSPARENCY:
                
                [self ReadTransparencyChunk  :(pc+len)  :[materialObject getTransparency]];
                break;
                
            case CHUNK_MAT_TWO_SIDE:
                
                [self Dsetpos  :(pc+len)];
                
                [materialObject setDoubleSided  :true];     // Unlikely, but handle if implemented
                break;

            case CHUNK_MAT_SHADING:
                
                [self Dsetpos  :(pc+len)];      // Currently unused
                break;
                
            case CHUNK_MAT_TEXMAP:
                
                [self ReadTextureOrBumpMapChunk  :(pc+len)  :[materialObject getTextureMap1]];
                break;

			case CHUNK_MAT_BUMPMAP:
                
				[self ReadTextureOrBumpMapChunk  :(pc+len)  :[materialObject getBumpMap]];
				break;
                
            case CHUNK_MAT_BUMP_PERCENT:
                
                [self Dsetpos  :(pc+len)];      // Currently unused
                break;
                
            case CHUNK_MAT_TEX2MAP:
                
                [self ReadTextureOrBumpMapChunk  :(pc+len)  :[materialObject getTextureMap2]];
                break;
                
			default: 
                
				[self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadMaterialBlocks: *** UNHANDLED IDENT *** 0x%X, length: %d", ident, len];
        }
    }
}

-(void) ReadMatNameChunk  :(P2Material*)mat {

    char  name[MAX_NAMESIZE] = {0};
    
    [self ReadCString  :name  :sizeof(name)];
    
	[mat setMatName  :name];

    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadMatChunk: '%s'", name];
}

-(void) ReadShadingChunk  :(int)p  :(P2RGBColor*)rgb {

    ushort  ident;
    int     len, pc;

    while ((pc = [self Dgetpos]) < p) {

		[self Dread  :&ident  :sizeof(ident)];
		[self Dread  :&len    :sizeof(len)];
		
		switch (ident) {

			case CHUNK_COLOR_F:
                
				[self ReadRGBFloat  :(pc+len)  :rgb];
				break;
                
			case CHUNK_COLOR_24:
                
				[self ReadRGBByte  :(pc+len)  :rgb];
				break;
                
			default:
                
				[self Dsetpos  :(pc+len)];
        }
    }
}

-(void) ReadShininessChunk  :(int)p  :(float*)shininess {

    ushort  ident;
    short   amount = 0;
    int     len, pc;

    while ((pc = [self Dgetpos]) < p) {

        [self Dread  :&ident  :sizeof(ident)];
        [self Dread  :&len    :sizeof(len)];
        
        switch (ident) {

            case CHUNK_INT_PERCENTAGE:
                
                [self ReadSignedShortChunk  :&amount];
                break;
                
            default:
                
                [self Dsetpos  :(pc+len)];
        }
    }
    
    *shininess = (float) amount;        // As read from the .P2 file, this is a 'percentage shininess'.  We'll scale
}                                       // this to an integer in the range [0..128] in the Convert_to_OpenGL routine

-(void) ReadTextureOrBumpMapChunk  :(int)p  :(P2TextureMap*)texture {

    ushort  ident;
    ushort  options = 0;
    short   percentage = 0;             // Texture, Bump or Texture2 percentage strength; currently read but unused
    char    name[MAX_NAMESIZE] = {0};
    int     len, pc;
    
    while ((pc = [self Dgetpos]) < p) {

		[self Dread  :&ident  :sizeof(ident)];
		[self Dread  :&len    :sizeof(len)];
        
		switch (ident) {

			case CHUNK_MAT_MAPNAME:
                
                [self ReadCString  :name  :sizeof(name)];
                
                [texture setTextureName  :name];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAPNAME: %s", [texture getTextureName]];
				break;
                
            case CHUNK_MAT_MAP_TILING:
                
                [self ReadUnsignedShortChunk  :&options];
                
                [texture setTextureOptions  :options];

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_TILING: 0x%X", options];
                break;

            case CHUNK_MAT_MAP_TEXBLUR:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_TEXBLUR (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_USCALE:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_USCALE (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_VSCALE:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_VSCALE (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_UOFFSET:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_UOFFSET (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_VOFFSET:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_VOFFSET (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_ANG:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_ANG (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_COL1:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_COL1 (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_COL2:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_COL2 (Unused)"];
                break;
                
            case CHUNK_MAT_MAP_RCOL:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_RCOL (Unused)"];
                break;

            case CHUNK_MAT_MAP_GCOL:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_GCOL (Unused)"];
                break;

            case CHUNK_MAT_MAP_BCOL:
                
                [self Dsetpos  :(pc+len)];                              // Currently unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_MAT_MAP_BCOL (Unused)"];
                break;
                
            case CHUNK_INT_PERCENTAGE:

                [self ReadSignedShortChunk  :&percentage];              // Currently read, but unused

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: CHUNK_INT_PERCENTAGE: %hi%% (Unused)", percentage];
                break;
                
            default:

				[self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadTextureOrBumpMapChunk: *** UNHANDLED IDENT *** 0x%X", ident];
        }
    }
}

-(void) ReadTransparencyChunk  :(int)p  :(float*)transparency {

    ushort  ident;
    short   percentage = 0;
    int     len, pc;
    
    while ((pc = [self Dgetpos]) < p) {

		[self Dread  :&ident  :sizeof(ident)];
		[self Dread  :&len    :sizeof(len)];
		
		switch (ident) {

			case CHUNK_INT_PERCENTAGE:
                
				[self ReadSignedShortChunk  :&percentage];
				break;
                
			default:
                
				[self Dsetpos  :(pc+len)];
        }
    }
	
    *transparency = (float) percentage;     // As read from the .P2 file, this is a 'percentage transparency'. We'll invert this
}                                           // to an 'alpha opacity' value needed by Convert_to_OpenGL routine

-(void) ReadKFDataBlocks  :(int)p {

    ushort  ident, animationLength;
    int     len, pc;

    while ((pc = [self Dgetpos]) < p) {

        [self Dread  :&ident  :sizeof(ident)];
        [self Dread  :&len    :sizeof(len)];
        
        switch (ident) {
                
            case CHUNK_KFHDR: {
                
                ushort  revision;
                char    filename[MAX_NAMESIZE] = {0};

                [self ReadUnsignedShortChunk  :&revision];          // Current keyframe section revision level
                [self ReadCString  :filename  :sizeof(filename)];   // Name of the original mesh file (Legacy)
                [self ReadUnsignedShortChunk  :&animationLength];   // Animation length in frames (1 to 32000)
                break;
            }
                
            case CHUNK_KFSEG: {

                ushort  firstFrame, lastFrame;

                [self ReadUnsignedShortChunk  :&firstFrame];        // First frame of animation segment
                [self ReadUnsignedShortChunk  :&lastFrame];         // Last  frame of animation segment
                break;
            }
                
            case CHUNK_KFCURTIME: {

                ushort  currentFrame;

                [self ReadUnsignedShortChunk  :&currentFrame];      // Current frame to render
                break;
            }
                
            case CHUNK_OBJECT_NODE_TAG: {

                [self ReadObjectNodeBlocks  :p  :animationLength];  // Object Node Blocks
                break;
            }

            default: {
                
                if ((len == 0) || (len == 32768)) {                 // Illegal, or misplaced block?
                    
                    [self Dsetpos  :p];                             // Bypass remaining KFDataBlock
                    
                    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadKFDataBlocks: *** ILLEGAL KFDataBlock BYPASSED *** Ident:0x%X, Length:%d", ident, len];
                    
                } else {
                    
                    [self Dsetpos  :(pc+len)];
                    
                    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadKFDataBlocks: *** UNHANDLED IDENT *** Ident:0x%X, Length: %d", ident, len];
                }
            }
        }
    }
}

-(void) ReadMasterScale  :(int)p {
    
    Vector3f *scaleVector = [Vector3f new];

    [self Dread  :[scaleVector getXptr]  :sizeof([scaleVector getX])];          // Read scale vector values
    [self Dread  :[scaleVector getYptr]  :sizeof([scaleVector getY])];
    [self Dread  :[scaleVector getZptr]  :sizeof([scaleVector getZ])];

    ScalingVector = scaleVector;                                                // Save scaling vector to scale bind pose transform, animation vertices

    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadMasterScale: CHUNK_MASTER_SCALE (%f, %f, %f)", [ScalingVector getX], [ScalingVector getY], [ScalingVector getZ]];
}

-(void) ReadObjectNodeBlocks  :(int)p  :(ushort)in_animLength {

//  Blender supports mesh object 'Actions' allowing animations without an Armature (i.e. bone) object. An example is a ray gun that fires light pulses. No bone involved.
//  Our custom Blender add-on (io_scene_p2) exports mesh object Actions by Category (position, rotation and scale) along with their Channels (X, Y and Z axes) containing
//  the collection of keyframe_points. Each keyframe_point is a 2-element vector of <frame number, value> with which to position/rotate/scale the mesh object's vertices.

    ushort          ident;
    int             len, pc;
    
    ushort          nodeID      = 0;                                            // Holds the last CHUNK_OBJECT_NODE_ID parsed
    NSMutableArray *meshObjects = [Scene getMeshObjList];                       // Get the listhead of valid P2MeshObj objects
    P2MeshObj      *meshObject  = nil;                                          // Holds P2MeshObj object pointed to by nodeID
    P2Action       *meshAction  = nil;                                          // Holds P2Action object of current P2MeshObj

    while ((pc = [self Dgetpos]) < p) {

        [self Dread  :&ident  :sizeof(ident)];
        [self Dread  :&len    :sizeof(len)];
        
        switch (ident) {
                
            case CHUNK_OBJECT_NODE_TAG: {                                       // Handle odd case of recursive NODE_TAG
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_OBJECT_NODE_TAG (Recursive)"]; // Log before recursive call

                [self ReadObjectNodeBlocks  :p  :in_animLength];                // Recursively handle Object Node Blocks
                break;
            }

            case CHUNK_OBJECT_NODE_ID: {                                        // nodeID (aka mesh object index number)

                [self ReadUnsignedShortChunk  :&nodeID];                        // nodeID chunk prefixes all *_NODE_TAGs

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_OBJECT_NODE_ID:%d", nodeID];
                break;
            }
                
            case CHUNK_NODE_HDR: {
                
                ushort  flags1, flags2, parent;                                 // (Node header flags aren't currently used)
                char    objectName[MAX_NAMESIZE] = {0};
                
                [self ReadCString  :objectName  :sizeof(objectName)];           // Object instance name; same as mesh's name
                [self ReadUnsignedShortChunk  :&flags1];                        // https://dzzd.net/3DSChunkDefinitions.html
                [self ReadUnsignedShortChunk  :&flags2];
                [self ReadUnsignedShortChunk  :&parent];                        // Parent nodeID, or 0xFFFF if not known
                
                if (nodeID < [meshObjects count]) {                             // NodeID limited to actual mesh objects
                                                                                // (i.e. ignore $$$DUMMY CHUNK_NODE_HDR)
                    meshObject = [meshObjects objectAtIndex  :nodeID];          // Get mesh object at current node index
                    
                    assert (strcmp (objectName, [meshObject getName]) == 0);    // Trap if we're not on the right object
                    
                    meshAction = [meshObject addAction];                        // Add Action to the current mesh object
                    
                    [meshAction setName             :objectName];               // Set the header values into the Action
                    [meshAction setNodeID           :nodeID];                   // Note: nodeID equals mesh object index
                    [meshAction setParentNodeID     :parent];                   // Set Action's parent node ID, or 0xFFFF
                    [meshAction setAnimationLength  :in_animLength];            // Propagate the animation length (frames)
                }
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_NODE_HDR '%s', nodeID:%d, flag1:%x, flag2:%x, parentID:%x",
                       objectName, nodeID, flags1, flags2, parent];
                break;
            }

            case CHUNK_INSTANCE_NAME: {                                         // Only seen for EMPTY chunks, so unused
                
                char    name[MAX_NAMESIZE] = {0};

                [self ReadCString  :name  :sizeof(name)];

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_INSTANCE_NAME '%s' (unused)", name];
                break;
            }

            case CHUNK_OBJECT_PARENT_NAME: {
                
                char    name[MAX_NAMESIZE] = {0};

                [self ReadCString  :name  :sizeof(name)];

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_OBJECT_PARENT_NAME '%s' (unused)", name];
                break;
            }

            case CHUNK_CAMERA_NODE_TAG: {

                [self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_CAMERA_NODE_TAG (Unused)"];
                break;
            }
                
            case CHUNK_TARGET_NODE_TAG: {

                [self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_TARGET_NODE_TAG (Unused)"];
                break;
            }
                
            case CHUNK_LIGHT_NODE_TAG: {

                [self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_LIGHT_NODE_TAG (Unused)"];
                break;
            }
                
            case CHUNK_LTARGET_NODE_TAG: {

                [self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_LTARGET_NODE_TAG (Unused)"];
                break;
            }
                
            case CHUNK_SPOT_NODE_TAG: {

                [self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_SPOT_NODE_TAG (Unused)"];
                break;
            }
                
            case CHUNK_OBJECT_PIVOT: {
                
                float   pivotX, pivotY, pivotZ;                                 // Rotational pivot point and scaling center
                
                [self Dread  :&pivotX  :sizeof(pivotX)];                        // Currently unused
                [self Dread  :&pivotY  :sizeof(pivotY)];
                [self Dread  :&pivotZ  :sizeof(pivotZ)];

                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_OBJECT_PIVOT x:%.03f y:%.03f z:%.03f", pivotX, pivotY, pivotZ];
                break;
            }

            case CHUNK_OBJECT_BOUNDBOX: {
                
                float   minBBx, minBBy, minBBz;                                 // Minimum point of mesh object bounding box
                float   maxBBx, maxBBy, maxBBz;                                 // Maximum point "   "     "       "      "
                
                [self Dread  :&minBBx  :sizeof(minBBx)];                        // Currently unused
                [self Dread  :&minBBy  :sizeof(minBBx)];
                [self Dread  :&minBBz  :sizeof(minBBz)];

                [self Dread  :&maxBBx  :sizeof(maxBBx)];
                [self Dread  :&maxBBy  :sizeof(maxBBy)];
                [self Dread  :&maxBBz  :sizeof(maxBBz)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_OBJECT_BOUNDBOX minimums: (%.03f, %.03f, %.03f), maximums: (%.03f, %.03f, %.03f)",
                       minBBx, minBBy, minBBz, maxBBx, maxBBy, maxBBz];
                break;
            }
                
            case CHUNK_POS_TRACK_TAG: {
                
                ushort    trackFlag, frameFlag;                                 // (Track, Frame flags are not currently used)
                uint32    frame_start, frame_end, nKeyframes, frameNum;
                float     kfPosition_X, kfPosition_Y, kfPosition_Z;
                
                [self Dread  :&trackFlag    :sizeof(trackFlag)];
                [self Dread  :&frame_start  :sizeof(frame_start)];
                [self Dread  :&frame_end    :sizeof(frame_end)];
                [self Dread  :&nKeyframes   :sizeof(nKeyframes)];
                
                if (meshAction) {                                               // Ignore $$$DUMMY mesh objects
                    
                    meshAction->actionLocation.frame_start = frame_start;       // Otherwise, save Position to Action header (as 'Location')
                    meshAction->actionLocation.frame_end   = frame_end;
                    meshAction->actionLocation.nKeys       = nKeyframes;
                }
                 
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_POS_TRACK_TAG nKeys:%d frameStart:%d frameEnd:%d", nKeyframes, frame_start, frame_end];

                for (uint32 index=0; index < nKeyframes; index++) {
                    
                    [self Dread  :&frameNum   :sizeof(frameNum)];               // Read keyframe number
                    [self Dread  :&frameFlag  :sizeof(frameFlag)];              // (Not currently used)

                    [self Dread  :&kfPosition_X   :sizeof(kfPosition_X)];       // Read keyframe position
                    [self Dread  :&kfPosition_Y   :sizeof(kfPosition_Y)];
                    [self Dread  :&kfPosition_Z   :sizeof(kfPosition_Z)];
                    
                    if (meshAction) {                                           // Ignore $$$DUMMY mesh objects

                        [meshAction->actionLocation.keyframes  setObject  :[[Vector3f alloc] init  :kfPosition_X  :kfPosition_Y  :kfPosition_Z]
                                                                    forKey:[[NSNumber alloc] initWithUnsignedInt  :frameNum]];
                    }

                    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_POS_TRACK_TAG frame:%d (%.03f, %.03f, %.03f)",
                           frameNum, kfPosition_X, kfPosition_Y, kfPosition_Z];
                }
                break;
            }
                
            case CHUNK_ROT_TRACK_TAG: {
                
                ushort  trackFlag, frameFlag;                                   // (Track, Frame flags are not currently used)
                uint32  frame_start, frame_end, nKeyframes, frameNum;
                float   kfAngle, kfAxis_X, kfAxis_Y, kfAxis_Z;
                
                [self Dread  :&trackFlag    :sizeof(trackFlag)];
                [self Dread  :&frame_start  :sizeof(frame_start)];
                [self Dread  :&frame_end    :sizeof(frame_end)];
                [self Dread  :&nKeyframes   :sizeof(nKeyframes)];
                
                if (meshAction) {                                               // Ignore $$$DUMMY mesh objects
                    
                    meshAction->actionRotation.frame_start = frame_start;       // Save Rotation to Action header
                    meshAction->actionRotation.frame_end   = frame_end;
                    meshAction->actionRotation.nKeys       = nKeyframes;
                }
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_ROT_TRACK_TAG nKeys:%d frameStart:%d frameEnd:%d", nKeyframes, frame_start, frame_end];

                for (uint32 index=0; index < nKeyframes; index++) {
                    
                    [self Dread  :&frameNum   :sizeof(frameNum)];               // Read keyframe number
                    [self Dread  :&frameFlag  :sizeof(frameFlag)];              // (Not currently used)

                    [self Dread  :&kfAngle   :sizeof(kfAngle)];                 // Read keyframe rotation angle (radians)
                    [self Dread  :&kfAxis_X  :sizeof(kfAxis_X)];                // Read keyframe rotation axes
                    [self Dread  :&kfAxis_Y  :sizeof(kfAxis_Y)];
                    [self Dread  :&kfAxis_Z  :sizeof(kfAxis_Z)];
                    
                    if (meshAction) {                                           // Ignore $$$DUMMY mesh objects

                        [meshAction->actionRotation.keyframes  setObject  :[[Vector4f alloc] init  :kfAxis_X  :kfAxis_Y  :kfAxis_Z  W:kfAngle]
                                                                    forKey:[[NSNumber alloc] initWithUnsignedInt  :frameNum]];
                    }

                    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_ROT_TRACK_TAG frame:%d (%.03f, %.03f, %.03f, w:%.03f)",
                           frameNum, kfAxis_X, kfAxis_Y, kfAxis_Z, kfAngle];
                }
                break;
            }
                
            case CHUNK_SCL_TRACK_TAG: {
                
                ushort  trackFlag, frameFlag;                                   // (Track, Frame flags are not currently used)
                uint32  frame_start, frame_end, nKeyframes, frameNum;
                float   kfScale_X, kfScale_Y, kfScale_Z;
                
                [self Dread  :&trackFlag    :sizeof(trackFlag)];
                [self Dread  :&frame_start  :sizeof(frame_start)];
                [self Dread  :&frame_end    :sizeof(frame_end)];
                [self Dread  :&nKeyframes   :sizeof(nKeyframes)];
                
                if (meshAction) {                                               // Ignore $$$DUMMY mesh objects
                    
                    meshAction->actionScale.frame_start = frame_start;          // Save Scale to Action header
                    meshAction->actionScale.frame_end   = frame_end;
                    meshAction->actionScale.nKeys       = nKeyframes;
                }
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_SCL_TRACK_TAG nKeys:%d frameStart:%d frameEnd:%d", nKeyframes, frame_start, frame_end];

                for (uint32 index=0; index < nKeyframes; index++) {
                    
                    [self Dread  :&frameNum   :sizeof(frameNum)];               // Read keyframe number
                    [self Dread  :&frameFlag  :sizeof(frameFlag)];              // (Not currently used)

                    [self Dread  :&kfScale_X  :sizeof(kfScale_X)];              // Read keyframe scale axes
                    [self Dread  :&kfScale_Y  :sizeof(kfScale_Y)];
                    [self Dread  :&kfScale_Z  :sizeof(kfScale_Z)];
                    
                    if (meshAction) {                                           // Ignore $$$DUMMY mesh objects     //TODO: Support scaling

                        [meshAction->actionScale.keyframes  setObject  :[[Vector3f alloc] init  :kfScale_X  :kfScale_Y  :kfScale_Z]
                                                                 forKey:[[NSNumber alloc] initWithUnsignedInt  :frameNum]];
                    }

                    [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: CHUNK_SCL_TRACK_TAG frame:%d (%.03f, %.03f, %.03f)",
                           frameNum, kfScale_X, kfScale_Y, kfScale_Z];
                }
                break;
            }

            default:

                [self Dsetpos  :(pc+len)];
                
                [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->ReadObjectNodeBlocks: *** UNHANDLED IDENT *** 0x%X", ident];
        }
    }
}

-(void) ReadSignedShortChunk  :(short*)amount {

    short temp;
    
    [self Dread	 :&temp  :sizeof(temp)];
    
    *amount = temp;
}

-(void) ReadUnsignedShortChunk  :(unsigned short*)amount {
    
    unsigned short temp;
    
    [self Dread  :&temp  :sizeof(temp)];
    
    *amount = temp;
}

-(void) ReadRGBFloat  :(int)p  :(P2RGBColor*)rgb {

    float red, green, blue;
    
    [self Dread  :&red    :sizeof(float)];
    [self Dread  :&green  :sizeof(float)];
    [self Dread  :&blue   :sizeof(float)];
    
    [rgb setRed    :(float) red];
    [rgb setGreen  :(float) green];
    [rgb setBlue   :(float) blue];
}

-(void) ReadRGBByte  :(int)p  :(P2RGBColor*)rgb {

    unsigned char red, green, blue;
    
    [self Dread  :&red    :sizeof(red)];
    [self Dread  :&green  :sizeof(green)];
    [self Dread  :&blue   :sizeof(blue)];
    
    [rgb setRed    :(float) (red   / 256.0)];
    [rgb setGreen  :(float) (green / 256.0)];
    [rgb setBlue   :(float) (blue  / 256.0)];
}

-(void) ReadCString  :(char *)outputCStr  :(uint)maxSize {

    uint    n = 0;
    
    do  {
        
        [self Dread  :&outputCStr[n++]  :1];
        
    } while (outputCStr[n-1] != '\0' && n < maxSize);
}

-(bool) LoadTextureFile  :(char*)szFileName  :(GLuint)TextureID {

    int				  rowNum, destRowNum = 0;
    unsigned char	 *flippedBytes;
    NSBitmapImageRep *image;
    
    if (!TextureID) return false;                                       // Pro Tip: Can't use glIsTexture here until glBindTexture's been called first

    NSString *path      = [MeshFileName stringByStandardizingPath];                                     // Get path including mesh filename
    NSString *purePath  = [[path stringByDeletingLastPathComponent] stringByAppendingString  :@"/"];    // Remove mesh filename and add back a traiing '/'
    
    NSString *texture   = [purePath stringByAppendingString  :[[NSString alloc] initWithCString  :szFileName encoding:NSUTF8StringEncoding]]; // Add texture filename to path

    if (!(image = (NSBitmapImageRep*)[NSBitmapImageRep imageRepWithContentsOfFile  :texture])) {
    
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->LoadTextureFile: NSBitmapImageRep imageRepWithContentsOfFile failed for TextureName: %@", texture];
        
        return false;                                                                   // Return as failed, leaving GL_TEXTURE_2D unbound
    }
    
    if ([image bitsPerPixel] != 32) {                                                   // Ignore non-RGBA textures; they're now crashing pre-Mojave systems

        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->LoadTextureFile: Ignoring texture with bitsPerPixel: %li, TextureName: %@", (long)[image bitsPerPixel], texture];

        return false;                                                                   // Return as failed, leaving GL_TEXTURE_2D unbound
    }
    
    if (!(flippedBytes = (unsigned char*)calloc ([image bytesPerRow] * [image pixelsHigh], 1))) {

        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->LoadTextureFile: calloc failed to allocate %ld bytes", [image bytesPerRow] * [image pixelsHigh]];

        return false;                                                                   // Return as failed, leaving GL_TEXTURE_2D unbound
    }
    
    for	(rowNum = (int)[image pixelsHigh] - 1; rowNum >= 0; rowNum--, destRowNum++) {   // Flip texture data from top to bottom (i.e. align with the OpenGL coordinate system)

		memcpy	(flippedBytes		+ (destRowNum * [image bytesPerRow]),
                 [image bitmapData]	+ (rowNum	  * [image bytesPerRow]),
                 [image bytesPerRow]);
    }

	glBindTexture	(GL_TEXTURE_2D, TextureID);
    
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    
	glTexParameteri	(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri	(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glTexImage2D    (GL_TEXTURE_2D, 0, GL_RGBA8, (int)[image pixelsWide], (int)[image pixelsHigh], 0, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8_REV, flippedBytes);
    
	free (flippedBytes);
	
	return true;
}

-(bool) Load1x1PixelTexture  :(GLuint)TextureID {

    NSBitmapImageRep *image;
    
    if (!TextureID) return false;                                       // Pro Tip: Can't use glIsTexture here until glBindTexture's been called first

    NSBundle *bundle = [NSBundle bundleForClass  :[self class]];

    NSString *pathToResource = [bundle pathForResource  :@"SinglePixelTexture" ofType:@"png"];

    if (!(image = (NSBitmapImageRep*)[NSBitmapImageRep imageRepWithContentsOfFile  :pathToResource])) {
    
        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->LoadTextureFile: NSBitmapImageRep imageRepWithContentsOfFile failed for texture:%@", pathToResource];
        
        return false;
    }
    
    if ([image bitsPerPixel] != 32) {                                   // Ignore non-RGBA texture

        [MeshP2 P2Log  :P2_Logging_Enabled  :@"MeshP2->LoadTextureFile: Ignoring texture with bitsPerPixel: %li, TextureName: %@", (long)[image bitsPerPixel], pathToResource];

        return false;
    }

    glBindTexture   (GL_TEXTURE_2D, TextureID);                         // Shader's sampler2D can't be empty (See https://stackoverflow.com/a/42239068/848364)

    glTexImage2D    (GL_TEXTURE_2D, 0, GL_RGBA8, (int)[image pixelsWide], (int)[image pixelsHigh], 0, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8_REV, [image bitmapData]);
    
    return true;
}

@end
