#import "Matrix4f.h"

@implementation Matrix4f            // A lot of this is ported from package org.lwjgl.util.vector

-(id) init {
    
    if ((self = [super init])) {
        
        m00 = 1.0f;  m01 = 0.0f;  m02 = 0.0f;  m03 = 0.0f;      // Initialize Column-major matrix with Identity
        m10 = 0.0f;  m11 = 1.0f;  m12 = 0.0f;  m13 = 0.0f;
        m20 = 0.0f;  m21 = 0.0f;  m22 = 1.0f;  m23 = 0.0f;
        m30 = 0.0f;  m31 = 0.0f;  m32 = 0.0f;  m33 = 1.0f;
    }
    
    return self;
}

-(Matrix4f*)    cmMatrixFromBytes  :(float*)sourceBytes {
    
    NSData      *data       = [NSData dataWithBytes  :sourceBytes length:64];
    NSUInteger  byteOffset  = 0;

    char        inputArray[sizeof(float)];
    float       outputArray[16];
    int32_t     orderedBytes;
    
    for (int i=0; i<16; i++) {

        [data getBytes  :inputArray range:NSMakeRange(byteOffset, sizeof(float))];
        
        byteOffset      += sizeof(float);
                                                    // (See https://www.debugcn.com/en/article/28269961.html)
        orderedBytes    = *(int32_t*)inputArray;    // Cast as a pointer to a 32-bit 'container' and copy 4 input bytes
        outputArray[i]  = *(float*)&orderedBytes;   // Cast the *address* of those 4 bytes to a float pointer and load them
    }
    
    m00 = outputArray[0];
    m01 = outputArray[1];
    m02 = outputArray[2];
    m03 = outputArray[3];

    m10 = outputArray[4];
    m11 = outputArray[5];
    m12 = outputArray[6];
    m13 = outputArray[7];

    m20 = outputArray[8];
    m21 = outputArray[9];
    m22 = outputArray[10];
    m23 = outputArray[11];

    m30 = outputArray[12];
    m31 = outputArray[13];
    m32 = outputArray[14];
    m33 = outputArray[15];

    return self;
}

-(Matrix4f*)    rmMatrixFromBytes  :(float*)sourceBytes {
    
    NSData      *data       = [NSData dataWithBytes  :sourceBytes length:64];
    NSUInteger  byteOffset  = 0;

    char        inputArray[sizeof(float)];
    float       outputArray[16];
    int32_t     orderedBytes;
    
    for (int i=0; i<16; i++) {

        [data getBytes  :inputArray range:NSMakeRange(byteOffset, sizeof(float))];
        
        byteOffset      += sizeof(float);
                                                    // (See https://www.debugcn.com/en/article/28269961.html)
        orderedBytes    = *(int32_t*)inputArray;    // Cast as a pointer to a 32-bit 'container' and copy 4 input bytes
        outputArray[i]  = *(float*)&orderedBytes;   // Cast the *address* of those 4 bytes to a float pointer and load them
    }

    m00 = outputArray[0];
    m10 = outputArray[1];
    m20 = outputArray[2];
    m30 = outputArray[3];

    m01 = outputArray[4];
    m11 = outputArray[5];
    m21 = outputArray[6];
    m31 = outputArray[7];

    m02 = outputArray[8];
    m12 = outputArray[9];
    m22 = outputArray[10];
    m32 = outputArray[11];

    m03 = outputArray[12];
    m13 = outputArray[13];
    m23 = outputArray[14];
    m33 = outputArray[15];

    return self;
}

-(Matrix4f*)    cmMatrixFromNSArray  :(NSArray*)sourceArray {

    m00 = [sourceArray[0] floatValue];
    m01 = [sourceArray[1] floatValue];
    m02 = [sourceArray[2] floatValue];
    m03 = [sourceArray[3] floatValue];

    m10 = [sourceArray[4] floatValue];
    m11 = [sourceArray[5] floatValue];
    m12 = [sourceArray[6] floatValue];
    m13 = [sourceArray[7] floatValue];

    m20 = [sourceArray[8] floatValue];
    m21 = [sourceArray[9] floatValue];
    m22 = [sourceArray[10] floatValue];
    m23 = [sourceArray[11] floatValue];

    m30 = [sourceArray[12] floatValue];
    m31 = [sourceArray[13] floatValue];
    m32 = [sourceArray[14] floatValue];
    m33 = [sourceArray[15] floatValue];

    return self;
}

-(Matrix4f*)    rmMatrixFromNSArray  :(NSArray*)sourceArray {

    m00 = [sourceArray[0] floatValue];
    m10 = [sourceArray[1] floatValue];
    m20 = [sourceArray[2] floatValue];
    m30 = [sourceArray[3] floatValue];

    m01 = [sourceArray[4] floatValue];
    m11 = [sourceArray[5] floatValue];
    m21 = [sourceArray[6] floatValue];
    m31 = [sourceArray[7] floatValue];

    m02 = [sourceArray[8] floatValue];
    m12 = [sourceArray[9] floatValue];
    m22 = [sourceArray[10] floatValue];
    m32 = [sourceArray[11] floatValue];

    m03 = [sourceArray[12] floatValue];
    m13 = [sourceArray[13] floatValue];
    m23 = [sourceArray[14] floatValue];
    m33 = [sourceArray[15] floatValue];

    return self;
}

-(float*)       getPtr { return &m00; }

+(Matrix4f*)    multiply  :(Matrix4f*)left  :(Matrix4f*)right {

    Matrix4f *dest = [Matrix4f new];

    float m00 = left->m00 * right->m00 + left->m10 * right->m01 + left->m20 * right->m02 + left->m30 * right->m03;
    float m01 = left->m01 * right->m00 + left->m11 * right->m01 + left->m21 * right->m02 + left->m31 * right->m03;
    float m02 = left->m02 * right->m00 + left->m12 * right->m01 + left->m22 * right->m02 + left->m32 * right->m03;
    float m03 = left->m03 * right->m00 + left->m13 * right->m01 + left->m23 * right->m02 + left->m33 * right->m03;
    float m10 = left->m00 * right->m10 + left->m10 * right->m11 + left->m20 * right->m12 + left->m30 * right->m13;
    float m11 = left->m01 * right->m10 + left->m11 * right->m11 + left->m21 * right->m12 + left->m31 * right->m13;
    float m12 = left->m02 * right->m10 + left->m12 * right->m11 + left->m22 * right->m12 + left->m32 * right->m13;
    float m13 = left->m03 * right->m10 + left->m13 * right->m11 + left->m23 * right->m12 + left->m33 * right->m13;
    float m20 = left->m00 * right->m20 + left->m10 * right->m21 + left->m20 * right->m22 + left->m30 * right->m23;
    float m21 = left->m01 * right->m20 + left->m11 * right->m21 + left->m21 * right->m22 + left->m31 * right->m23;
    float m22 = left->m02 * right->m20 + left->m12 * right->m21 + left->m22 * right->m22 + left->m32 * right->m23;
    float m23 = left->m03 * right->m20 + left->m13 * right->m21 + left->m23 * right->m22 + left->m33 * right->m23;
    float m30 = left->m00 * right->m30 + left->m10 * right->m31 + left->m20 * right->m32 + left->m30 * right->m33;
    float m31 = left->m01 * right->m30 + left->m11 * right->m31 + left->m21 * right->m32 + left->m31 * right->m33;
    float m32 = left->m02 * right->m30 + left->m12 * right->m31 + left->m22 * right->m32 + left->m32 * right->m33;
    float m33 = left->m03 * right->m30 + left->m13 * right->m31 + left->m23 * right->m32 + left->m33 * right->m33;
    
    dest->m00 = m00;
    dest->m01 = m01;
    dest->m02 = m02;
    dest->m03 = m03;
    dest->m10 = m10;
    dest->m11 = m11;
    dest->m12 = m12;
    dest->m13 = m13;
    dest->m20 = m20;
    dest->m21 = m21;
    dest->m22 = m22;
    dest->m23 = m23;
    dest->m30 = m30;
    dest->m31 = m31;
    dest->m32 = m32;
    dest->m33 = m33;
    
    return dest;
}

+(Matrix4f*)    scale  :(Vector3f*)in_vec  :(Matrix4f*)in_src {
    
    glm::vec3 glmVec3 = glm::vec3 ([in_vec getX], [in_vec getY], [in_vec getZ]);
    
    glm::mat4 glmMat4 = glm::make_mat4 ([in_src getPtr]);

              glmMat4 = glm::scale (glmMat4, glmVec3);
    
    return [[Matrix4f new] rmMatrixFromBytes  :(float*)glm::value_ptr (glmMat4)];
}

+(Matrix4f*)    translate  :(Vector3f*)vec  :(Matrix4f*)src {
    
    Matrix4f *dest = [Matrix4f new];

    dest->m30 += (src->m00 * [vec getX]) + (src->m10 * [vec getY]) + (src->m20 * [vec getZ]);
    dest->m31 += (src->m01 * [vec getX]) + (src->m11 * [vec getY]) + (src->m21 * [vec getZ]);
    dest->m32 += (src->m02 * [vec getX]) + (src->m12 * [vec getY]) + (src->m22 * [vec getZ]);
    dest->m33 += (src->m03 * [vec getX]) + (src->m13 * [vec getY]) + (src->m23 * [vec getZ]);
    
    return dest;
}

+(Matrix4f*)    transpose  :(Matrix4f*)src {
    
    Matrix4f *dest = [Matrix4f new];

    float m00 = src->m00;
    float m01 = src->m10;
    float m02 = src->m20;
    float m03 = src->m30;
    float m10 = src->m01;
    float m11 = src->m11;
    float m12 = src->m21;
    float m13 = src->m31;
    float m20 = src->m02;
    float m21 = src->m12;
    float m22 = src->m22;
    float m23 = src->m32;
    float m30 = src->m03;
    float m31 = src->m13;
    float m32 = src->m23;
    float m33 = src->m33;
    
    dest->m00 = m00;
    dest->m01 = m01;
    dest->m02 = m02;
    dest->m03 = m03;
    dest->m10 = m10;
    dest->m11 = m11;
    dest->m12 = m12;
    dest->m13 = m13;
    dest->m20 = m20;
    dest->m21 = m21;
    dest->m22 = m22;
    dest->m23 = m23;
    dest->m30 = m30;
    dest->m31 = m31;
    dest->m32 = m32;
    dest->m33 = m33;
    
    return dest;
}

@end
