#import "P2Action.h"
#import "Categories.h"

@implementation P2Action

-(id) init {
    
    if (self = [super init]) {
        
        actionName[0]                = 0;
        actionNodeID                 = 0;
        actionParentNodeID           = 0;
        actionMsBindTransform        = [Matrix4f new];
        actionAnimations             = [NSArray  new];
        actionAnimationLength        = 0;
        actionLocation.frame_start   = 0;
        actionLocation.frame_end     = 0;
        actionLocation.nKeys         = 0;
        actionLocation.keyframes     = [NSMutableDictionary<NSNumber*, Vector3f*> new];
        actionRotation.frame_start   = 0;
        actionRotation.frame_end     = 0;
        actionRotation.nKeys         = 0;
        actionRotation.keyframes     = [NSMutableDictionary<NSNumber*, Vector4f*> new];
        actionScale.frame_start      = 0;
        actionScale.frame_end        = 0;
        actionScale.nKeys            = 0;
        actionScale.keyframes        = [NSMutableDictionary<NSNumber*, Vector3f*> new];
    }
    
    return self;
}

-(char*)     getName            { return actionName; }
-(ushort)    getNodeID          { return actionNodeID; }
-(ushort)    getParentNodeID    { return actionParentNodeID; }
-(Matrix4f*) getMsBindTransform { return actionMsBindTransform; }
-(NSArray*)  getAnimations      { return actionAnimations; }
-(ushort)    getAnimationLength { return actionAnimationLength; }

-(void) setName             :(char*)new_Name                 { strcpy (actionName, new_Name); }
-(void) setNodeID           :(ushort)new_NodeID              { actionNodeID = new_NodeID; }
-(void) setParentNodeID     :(ushort)new_ParentNodeID        { actionParentNodeID = new_ParentNodeID; }
-(void) setMsBindTransform  :(Matrix4f*)new_MsBindTransform  { actionMsBindTransform = new_MsBindTransform; }
-(void) setAnimationLength  :(ushort)new_AnimationLength     { actionAnimationLength = new_AnimationLength; }

-(void) synthesizeActionAnimations {

    //  Create an array of P2Action animation transforms, similar to P2ArmatureBone Blender-imported animation transforms

    __block NSMutableArray *animations = [NSMutableArray new];

    NSArray *sortedKeyframes = [actionLocation.keyframes sortedKeys];           // Locations, sorted by keyframe time (seconds)

    [sortedKeyframes enumerateObjectsUsingBlock  :^(id object, NSUInteger idx, BOOL *stop) {
        
        Vector3f   *position        = [actionLocation.keyframes objectForKey  :object];
        Vector4f   *rotation        = [actionRotation.keyframes objectForKey  :object];
        Vector3f   *scale           = [   actionScale.keyframes objectForKey  :object];

        glm::vec3   glmPosition     = glm::vec3 ([position getX], [position getY], [position getZ]);    // Get Model-space position, rotation and scale
        glm::vec3   glmRotation     = glm::vec3 ([rotation getX], [rotation getY], [rotation getZ]);
        glm::vec3   glmScale        = glm::vec3 (   [scale getX],    [scale getY],    [scale getZ]);

        float       rotAngle        = [rotation getW];

        glm::mat4   glmTransform    = glm::make_mat4 ([actionMsBindTransform getPtr]);  // Get P2Action's Model-space animation bind transform
        
        glmTransform = glm::translate (glmTransform,           glmPosition);    // Follow OpenGL's familiar 'TRS' ordering: Translate, Rotate and Scale
        glmTransform = glm::rotate    (glmTransform, rotAngle, glmRotation);
        glmTransform = glm::scale     (glmTransform, glmScale);                 // Note: Scale is applied to transform, but DOESN'T SURVIVE CONVERSION TO DQ!

        [animations addObject  :[[Matrix4f new] cmMatrixFromBytes  :glm::value_ptr (glmTransform)]]; // Add Column-major matrix to animation transforms array
    }];
    
    actionAnimations = [animations copy];                                       // Immutable copy of animation transforms array to P2Action instance variable
}

@end
