#import "Constants.h"
#import <OpenGL/gl3.h>

@interface POLYGONLIST : NSObject {
	
    char		name[MAX_NAMESIZE];			    // Material name
    char        textureMap1Name[MAX_NAMESIZE];  // 1st TextureMap name
    char        textureMap2Name[MAX_NAMESIZE];  // 2nd TextureMap name
    char        bumpMapName[MAX_NAMESIZE];      // BumpMap name
        
    GLfloat		ambient[4];
    GLfloat		diffuse[4];
    GLfloat		specular[4];
    GLfloat     shininess;
    GLfloat		alpha;

    BOOL		doublesided;
        
    GLuint      textureMap1ID;
    GLuint      textureMap2ID;
    GLuint      bumpMapID;

    GLuint      materialIBO;                // IBO indexes into the VBO for this material's vertices, normals and texcoords

    GLuint      *iboBuffer;                 // Client-side IBO buffer, an index of Points (to surfaces using this material)
    GLuint      iboSize;
        
    NSMutableArray	*SurfaceIndexList;		// Array of SURFACEs, whose Points (P0, P1, P2) are indices into VertexList
}

-(id)		getSurfaceIndexList;
-(int)		getSurfaceIndexListCount;

-(id)		addSurfaceIndexList;

-(char*)	getName;
-(char*)    getTextureMap1Name;
-(char*)    getTextureMap2Name;
-(char*)    getBumpMapName;

-(void)     genTextureMap1ID;
-(void)     genTextureMap2ID;
-(void)     genBumpMapID;
-(void)     genMaterialIBO;

-(GLuint)   getTextureMap1ID;
-(GLuint)   getTextureMap2ID;
-(GLuint)   getBumpMapID;
-(GLuint)   getMaterialIBO;

-(GLuint*)  getIBOBuffer;
-(GLuint)   getIBOSize;

-(GLfloat*)	getAmbient;
-(GLfloat*)	getDiffuse;
-(GLfloat*)	getSpecular;
-(GLfloat)  getShininess;
-(GLfloat)  getAlpha;

-(BOOL)		getDoubleSided;

-(void)		setName			    :(char*)new_Name;
-(void)     setTextureMap1Name  :(char*)new_Texture1Name;
-(void)     setTextureMap2Name  :(char*)new_Texture2Name;
-(void)     setBumpMapName      :(char*)new_BumpMapName;
-(void)		setAmbient 		    :(GLfloat*)new_Ambient;
-(void)		setDiffuse 		    :(GLfloat*)new_Diffuse;
-(void)		setSpecular		    :(GLfloat*)new_Specular;
-(void)     setShininess        :(GLfloat)new_Shininess;
-(void)		setAlpha            :(GLfloat)new_Alpha;
-(void)		setDoubleSided	    :(BOOL)new_DoubleSided;
-(void)     setIBOBuffer        :(GLuint*)new_iboBuffer;
-(void)     setIBOSize          :(GLuint)new_iboSize;

@end
