#import "Quaternion.h"

@implementation Quaternion

-(id) init  :(float)in_X  :(float)in_Y  :(float)in_Z  W:(float)in_W {
    
    if ((self = [super init])) {
        
        x = in_X; y = in_Y; z = in_Z; w = in_W;
    }
    
    return self;
}

-(id) initFromAngleAxis  :(float)angleDegrees  :(Vector3f*)axes {
    
    if ((self = [super init])) {                // From: https://stackoverflow.com/a/6281055
        
        [axes normalize];
        
        float theta = glm::radians (angleDegrees);
        
        x = sin (theta*0.5) * [axes getX];      // Rotation X-axis (normalized)
        y = sin (theta*0.5) * [axes getY];      //      "   Y-axis      "
        z = sin (theta*0.5) * [axes getZ];      //      "   Z-axis      "
        w = cos (theta*0.5);                    // Rotation angle  (in radians)
    }
    
    return self;
}

-(float)  getX    { return x; }
-(float)  getY    { return y; }
-(float)  getZ    { return z; }
-(float)  getW    { return w; }
-(float*) getXptr { return &x; }

-(NSString*) toString   { return [NSString stringWithFormat  :@"(%f, %f, %f, W:%f)", x, y, z, w]; }

+(Quaternion*) normalize  :(Quaternion*)in_Quaternion {
    
    assert (in_Quaternion != nil);
    
    glm::quat glmQuaternion = glm::quat (in_Quaternion.getW, in_Quaternion.getX, in_Quaternion.getY, in_Quaternion.getZ);
    
    glm::quat glmNormalized = glm::normalize (glmQuaternion);
    
    return [[Quaternion alloc] init  :glmNormalized.x  :glmNormalized.y  :glmNormalized.z  W:glmNormalized.w];;
}

+(Quaternion*) multiply  :(Quaternion*)lhs  :(Quaternion*)rhs {     // From https://donw.io/post/dual-quaternion-skinning/

    glm::quat glmProduct = [Quaternion toGLMquat  :lhs] * [Quaternion toGLMquat  :rhs];
    
    return [[Quaternion alloc] init  :glmProduct.x :glmProduct.y :glmProduct.z W:glmProduct.w];
}

+(Quaternion*) interpolate  :(Quaternion*)fromQ  :(Quaternion*)toQ  :(float)progression {
    
    assert ((fromQ != nil) && (toQ != nil));
    assert ((progression >= 0.0) && (progression <= 1.0));

    glm::quat glmLERP = glm::lerp ([self toGLMquat  :fromQ], [self toGLMquat  :toQ], progression);
    
    return [[Quaternion alloc] init  :glmLERP.x  :glmLERP.y  :glmLERP.z  W:glmLERP.w];
}

+(glm::quat) toGLMquat  :(Quaternion*)q {
    
    assert (q != nil);
  
    return glm::quat (q.getW,  q.getX,  q.getY,  q.getZ);   // Ordering is glm::quat (W,X,Y,Z) .vs. glm::vec4 (x,y,z,w)
}

@end
