#import "VERTEX.h"

@implementation VERTEX

-(id) init {
	
    if ((self = [super init])) {
        
        x = y = z = nx = ny = nz = tx = ty = tz = u = v = 0.0;
        
        flags                 = [NSMutableArray new];
        boneIndicesAndWeights = [NSArray new];
    }
    
    return self;
}

-(id) init  :(GLfloat)new_X     :(GLfloat)new_Y     :(GLfloat)new_Z
            :(GLfloat)new_NX    :(GLfloat)new_NY    :(GLfloat)new_NZ
            :(GLfloat)new_TX    :(GLfloat)new_TY    :(GLfloat)new_TZ
            :(GLfloat)new_U     :(GLfloat)new_V
            :(NSMutableArray*)new_Flags {
	
    if ((self = [super init])) {
		
        x     = new_X;	 y  = new_Y;	z  = new_Z;
		nx    = new_NX;  ny = new_NY;	nz = new_NZ;
        tx    = new_TX;  ty = new_TY;   tz = new_TZ;
        u     = new_U;   v  = new_V;

        flags                 = new_Flags;
        boneIndicesAndWeights = [NSArray new];;
    }
    
    return self;
}
    
-(GLfloat)  getX    { return x; }
-(GLfloat)  getY    { return y; }
-(GLfloat)  getZ    { return z; }

-(GLfloat)  getNX	{ return nx; }
-(GLfloat)  getNY	{ return ny; }
-(GLfloat)  getNZ	{ return nz; }

-(GLfloat)  getTX	{ return tx; }
-(GLfloat)  getTY	{ return ty; }
-(GLfloat)  getTZ	{ return tz; }

-(GLfloat)  getU     { return u; }
-(GLfloat)  getV     { return v; }

-(NSMutableArray*)  getFlags                 { return flags; }
-(NSArray*)         getBoneIndicesAndWeights { return boneIndicesAndWeights; }

-(void)	    setX	  :(GLfloat)new_X  { x=new_X; }
-(void)	    setY	  :(GLfloat)new_Y  { y=new_Y; }
-(void)	    setZ	  :(GLfloat)new_Z  { z=new_Z; }

-(void)	    setNX	  :(GLfloat)new_NX { nx=new_NX; }
-(void)	    setNY	  :(GLfloat)new_NY { ny=new_NY; }
-(void)	    setNZ	  :(GLfloat)new_NZ { nz=new_NZ; }

-(void)	    setTX	  :(GLfloat)new_TX { tx=new_TX; }
-(void)     setTY     :(GLfloat)new_TY { ty=new_TY; }
-(void)     setTZ     :(GLfloat)new_TZ { tz=new_TZ; }

-(void)     setU      :(GLfloat)new_U  { u=new_U; }
-(void)     setV      :(GLfloat)new_V  { v=new_V; }

-(void)     addFlag   :(GLushort)new_Flag         { [flags addObject  :[NSNumber numberWithUnsignedShort  :new_Flag]]; }
-(void)     setFlags  :(NSMutableArray*)new_Flags { flags = new_Flags; }

-(void)	    Zoom	  :(float)zoom { x *= zoom; y *= zoom; z *= zoom; }

-(void)     BuildBoneIndicesAndWeightsArray {
    
    uint16           flag;
    uint16           boneIndex;
    uint8            boneWeight;
    NSMutableArray  *stagingArray = [[NSMutableArray alloc] initWithCapacity  :MAX_WEIGHTS];
    
    for (int i=0; i < MAX_WEIGHTS; i++) {
        
        @try {                                                              // Add Index/Weight or a placeholder

            flag         = [flags[i] unsignedShortValue];                   // Dereference the vertex flag element
            
            boneIndex    = ((flag & 0xFF80) >> 7);                          // Isolate Bone Index (9 bits from 7-15)
            boneWeight   =  (flag & 0x007F);                                // Isolate Vertex Weight (7 bits from 0-6)
        
        } @catch (NSException *exception) {                                 // Trap if no flag at this flags array slot
            
            boneIndex    = MAX_BONES;                                       // Flag index with out-of-range placeholder
            boneWeight   = 0;                                               // With a weight indicating 'no deformation'
        
        } @finally {                                                        // Add to staging array
            
            NSNumber        *boneNumber    = [[NSNumber alloc] initWithUnsignedInt  :boneIndex];
            NSNumber        *weightNumber  = [[NSNumber alloc] initWithFloat  :(boneWeight / 100.0)];  // Convert from %
            
            NSMutableArray  *unsortedArray = [[NSMutableArray alloc] initWithObjects  :boneNumber, weightNumber, nil];
            
            [stagingArray addObject  :unsortedArray];
        }
    }

    [stagingArray sortUsingComparator: ^(id a, id b) {                      // Sort staging array by bone weight

        NSNumber *numA = a[1];                                              // Sort by 2nd element (i.e. 'weight')
        NSNumber *numB = b[1];
        
        return [numB compare  :numA];                                       // Descending order (highest to lowest)
    }];

    normalizeByWeight (stagingArray);
    
    boneIndicesAndWeights = stagingArray;
}

static void normalizeByWeight (NSMutableArray* in_SortedByWeightArray) {
    
    float sum   = 0.0;
    float value = 0.0;
    
    for (int i=0; i < [in_SortedByWeightArray count]; i++) {    // For all array elements, sorted by weight
        
        if (i == MAX_WEIGHTS) break;                            // Limit to only the top MAX_WEIGHTS elements
        
        value  = [in_SortedByWeightArray[i][1] floatValue];

        if (value > 1.0) value = 1.0;                           // Clamp at maximum weight

        sum += value;                                           // Accumulate total weight entries
    }

    for (int i=0; i < [in_SortedByWeightArray count]; i++) {    // Re-scan array and write normalized values
        
        value  = [in_SortedByWeightArray[i][1] floatValue];
        
        if (value > 1.0) value = 1.0;                           // Clamp at maximum weight

        if (sum != 0.0) {                                       // Bypass divide-by-zero errors
            
            value /= sum;                                       // Looks okay - compute normalized value
        
        } else value = 0.0;                                     // Otherwise, default to nulled out value

        [in_SortedByWeightArray[i] replaceObjectAtIndex  :1 withObject:[NSNumber numberWithFloat  :value]];
    }
}

@end
