#import "Vector3f.h"

@implementation Vector3f

-(id)   init  :(float)in_X   :(float)in_Y   :(float)in_Z {
    
    if ((self = [super init])) {
        
        x   = in_X;
        y   = in_Y;
        z   = in_Z;
    }
    
    return self;
}

-(id)   initFromGLM  :(glm::vec3)in_glmVec3 {
    
    if ((self = [super init])) {
        
        x   = in_glmVec3.x;
        y   = in_glmVec3.y;
        z   = in_glmVec3.z;
    }
    
    return self;
}

-(float)    getX    { return x; }
-(float)    getY    { return y; }
-(float)    getZ    { return z; }

-(float*)   getXptr { return &x; }
-(float*)   getYptr { return &y; }
-(float*)   getZptr { return &z; }

-(void)     setX    :(float)in_X { x = in_X; };
-(void)     setY    :(float)in_Y { y = in_Y; };
-(void)     setZ    :(float)in_Z { z = in_Z; };

-(id)       normalize {
    
    float mag = (x*x) + (y*y) + (z*z);
    
    if (mag < FLT_EPSILON) return self;
    
    float invLength = 1.0f / sqrt (mag);
    
    x *= invLength;
    y *= invLength;
    z *= invLength;

    return self;
}

+(Vector3f*)    interpolateVector  :(Vector3f*)start  :(Vector3f*)end  :(float)progression {

    float x = [start getX] + (([end getX] - [start getX]) * progression);
    float y = [start getY] + (([end getY] - [start getY]) * progression);
    float z = [start getZ] + (([end getZ] - [start getZ]) * progression);

    return [[Vector3f alloc] init  :x  :y  :z];
}

+(Vector3f*)    scale  :(Vector3f*)in_vec  :(float)in_scale {
    
    glm::vec3 glmVec3 = in_scale * glm::vec3 ([in_vec getX], [in_vec getY], [in_vec getZ]);
    
    return [[Vector3f alloc] init  :glmVec3.x  :glmVec3.y  :glmVec3.z];
}

+(Vector3f*)    add  :(Vector3f*)left  :(Vector3f*)right {
    
    return [[Vector3f alloc] init  :(left.getX + right.getX)  :(left.getY + right.getY)  :(left.getZ + right.getZ)];
}

+(Vector3f*)    sub  :(Vector3f*)left  :(Vector3f*)right {
    
    return [[Vector3f alloc] init  :(left.getX - right.getX)  :(left.getY - right.getY)  :(left.getZ - right.getZ)];
}

+(Vector3f*)    multiply  :(Vector3f*)left  :(Vector3f*)right {
    
    return [[Vector3f alloc] init  :(left.getX * right.getX)  :(left.getY * right.getY)  :(left.getZ * right.getZ)];
}

+(Vector3f*)    cross  :(Vector3f*)v1  :(Vector3f*)v2 {

    Vector3f *dest = [Vector3f new];

    [dest setX  :(v1.getY * v2.getZ - v1.getZ * v2.getY)];
    [dest setY  :(v2.getX * v1.getZ - v2.getZ * v1.getX)];
    [dest setZ  :(v1.getX * v2.getY - v1.getY * v2.getX)];
    
    return dest;
}

+(float)        dot  :(Vector3f*)left  :(Vector3f*)right {
    
    return ([left getX] * [right getX]) + ([left getY] * [right getY]) + ([left getZ] * [right getZ]);
}

@end
